"""
aiodec
======

Decorators for coroutines

"""

import time
import logging
from functools import wraps
from string import Template
import inspect
from inspect import Signature
from typing import Callable, Optional, Mapping, Any

__version__ = '2018.10.1'

logger = logging.getLogger(__name__)
Callback = Callable[[Signature, Mapping[str, Any]], None]


def adecorator(
        f=None,
        pre_callback: Optional[Callback] = None,
        post_callback: Optional[Callback]= None):

    def inner(g):

        @wraps(g)
        async def wrapper(*args, **kwargs):
            # Get the function signature of the wrapped function. We need this
            # in order to obtain all the parameter information.
            template_parameters = dict(name_=g.__name__, qualname_=g.__qualname__)
            sig = inspect.signature(g)

            # Using the actually-provided args, bind them to the signature
            # of the wrapped function
            bound_args = sig.bind(*args, **kwargs)
            
            # Now fill in the unsupplied parameters with their default
            # values.
            bound_args.apply_defaults()
            template_parameters.update(bound_args.arguments)
            pre_callback and pre_callback(sig, template_parameters)
            try:
                return await g(*args, **kwargs)
            finally:
                post_callback and post_callback(sig, template_parameters)

        return wrapper

    if f:
        # astopwatch() was called WITHOUT evaluation, so we must return the
        # actual wrapper function that replaces f.
        return inner(f)
    else:
        # astopwatch() was called WITH evaluation, so we need to return ANOTHER
        # decorator (that will receive  and wrap function f).
        return inner


def astopwatch(
        f=None,
        message_template='Time taken: $time_ seconds',
        fmt='%.4g',
        logger=logger,
):
    # Using templates because safe_substitute is awesome.
    tmpl = Template(message_template)
    t0 = 0

    def pre_callback(sig, template_parameters):
        nonlocal t0
        t0 = time.perf_counter()

    def post_callback(sig, template_parameters):
        nonlocal t0
        dt = time.perf_counter() - t0
        msg = tmpl.safe_substitute(
            **template_parameters,
            time_=fmt % dt
        )
        logger.info(msg)

    return adecorator(
        f,
        pre_callback=pre_callback,
        post_callback=post_callback
    )



import logging
import asyncio
import pytest
from aiodec import astopwatch


logging.basicConfig(level='DEBUG')


@pytest.fixture
def loop():
    return asyncio.get_event_loop()


def test_timer_default(loop):

    @astopwatch
    async def f(x):
        await asyncio.sleep(0.01)
        return x + 1

    result = loop.run_until_complete(f(3))
    assert result == 4


def test_timer_custom(loop):

    @astopwatch(message_template='Blah $_time s')
    async def f(x):
        await asyncio.sleep(0.01)
        return x + 1

    result = loop.run_until_complete(f(3))
    assert result == 4


def test_timer_invalid_template(loop):

    @astopwatch(message_template='Blah')
    async def f(x):
        await asyncio.sleep(0.01)
        return x + 1

    result = loop.run_until_complete(f(3))
    assert result == 4


def test_timer_invalid_raise(loop):

    @astopwatch(message_template='RuntimeError test took: $time_')
    async def f():
        await asyncio.sleep(0.01)
        raise RuntimeError('blah')

    with pytest.raises(RuntimeError):
        loop.run_until_complete(f())


def test_timer_moar_template(loop, caplog):

    # Observe: the message template is going to include parameter
    # values from the wrapped function.
    @astopwatch(message_template='fn=$name_ abc=$abc ddd=$ddd time=$time_ secs')
    async def f(abc, ddd=123):
        await asyncio.sleep(0.01)
        return 'ok'

    result = loop.run_until_complete(f(3))
    assert result == 'ok'
    expected_fragment = 'fn=f abc=3 ddd=123 time='
    assert any(expected_fragment in r.message for r in caplog.records)  # pragma: no cover



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

extras_require = \
{'dev': ['pytest', 'pytest-cov', 'wheel', 'coveralls']}

setup(name='aiodec',
      version='2018.10.1',
      description='aiodec',
      author='Caleb Hattingh',
      author_email='caleb.hattingh@gmail.com',
      url='https://github.com/cjrh/aiodec',
      py_modules=['aiodec'],
      extras_require=extras_require,
     )
